'''
Defines the Publishable class and the module level variable
to_publish.
'''
to_publish = []
'''
This variable is intented to be used only in the ryzom.apps
AppConfig.ready() to populate the DB with publications
at server startup.
'''


class Publishable:
    '''
    The publishable class is meant to be inherited from
    by end user models. It defines a publish class method
    that's used to create a publication associated with the
    model inherited from it.
    '''
    _published = False
    _prepubs = {}

    @classmethod
    def publish(cls, name, template=None):
        '''
        This method permit the publication of a model
        specifying the (unique) name of the publication,
        the template used to render the model inheriting from
        this class, and the query that will filter, order,
        limit, ..etc, the set published.
        It defers the publication until the DB is ready to
        accept new entries.

        :param str name: A unique name
        :param str template: A Component module and class \
                `module.submodule.Class`
        :param list query: A list of dict of query parameters

        :examples:
            ::

                Task.publish('task', 'todos.components.task.Task', [
                    {'order_by': '-about'},
                    {'limit': 5},
                    {'offset': 3}
                ])
        '''
        if not cls._published:
            cls._prepubs[name] = template
            if cls not in to_publish:
                to_publish.append(cls)
        else:
            cls.do_publish(name, template)

    @classmethod
    def do_publish(cls, name, template):
        '''f
        This method actually created the real publication,
        called by the publish_all() method, or by publish()
        if the publishable has already been published.
        If a publication of given name already exists,
        this method only updates its fields.

        :parameters: see :meth:`publish`
        '''
        from ryzom.models import Publication

        tmpl_mod, tmpl_cls = template.rsplit('.', 1)
        kwargs = {
            'model_module': cls.__module__,
            'model_class': cls.__name__,
            'template_module': tmpl_mod,
            'template_class': tmpl_cls,
        }
        pub_exists = Publication.objects.filter(name=name).exists()
        if pub_exists:
            pub = Publication.objects.get(name=name)
            changed = False
            for k, v in kwargs.items():
                if getattr(pub, k, None) != v:
                    setattr(pub, k, v)
                    changed = True
            if changed:
                pub.save()
        else:
            Publication.objects.create(name=name, **kwargs)

    @classmethod
    def publish_all(cls):
        '''
        This method is called by the ryzom.apps AppConfig.ready()
        at server startup to populate the DB with the publications
        associated to the current model class.
        '''
        cls._published = True
        for k, v in cls._prepubs.items():
            name, template = k, v
            cls.do_publish(name, template)


def publish(component):
    def func_wrapper(func):
        @classmethod
        def wrapper(*args):
            cls = args[0]
            cls.publish(func.__name__, component)
            return func(*args)
        return wrapper
    return func_wrapper



from django.core.asgi import get_asgi_application
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.auth import AuthMiddlewareStack

import ryzom.routing

application = ProtocolTypeRouter({
    'http': get_asgi_application(),
    'websocket': AuthMiddlewareStack(
        URLRouter(
            ryzom.routing.websocket_urlpatterns
        )
    ),
})



from django.conf import settings

from ryzom.components import Div, H1, A


class Welcome(Div):
    def __init__(self):
        content = [
            H1('WELCOME !'),
            A('Go to TODO list', **{'href': 'todos'}),
        ]
        super().__init__(*content)



from django.template.loader import render_to_string
from jinja2.utils import contextfunction


@contextfunction
def render_form(context, form):
    # django-betterforms has multiple calls for different forms so
    # using context['parent']['view'] isn't simple.
    # Jinja2 Context is immutable, so create our own with form parameter.
    context = dict(context=context,
                   form=form,
                   )
    return render_to_string('ryzom.components.django.Form', context)



from django.utils.safestring import mark_safe
from ryzom.html import *


def component_html(path, *args, **kwargs):
    try:
        from jinja2.utils import Markup
    except ImportError:
        Markup = None
    # todo: replace with importlib
    import cli2
    ComponentCls = cli2.Node(path).target
    component = ComponentCls(*args, **kwargs)
    html = component.to_html()

    if Markup:
        html = Markup(html)
    return mark_safe(html)



from crudlfap import shortcuts as crudlfap

from .models import Task


class TaskMixin:
    allowed_groups = 'any'

    def get_exclude(self):
        req = getattr(self, 'request', None)
        if (req is None
            or
                req and not self.request.user.is_staff):
            return ['user']
        return super().get_exclude()


class TaskCreateView(TaskMixin, crudlfap.CreateView):
    def form_valid(self):
        self.form.instance.user = self.request.user
        return super().form_valid()


class TaskRouter(crudlfap.Router):
    fields = '__all__'
    material_icon = 'view_list'
    model = Task

    views = [
        crudlfap.DeleteView.clone(TaskMixin),
        crudlfap.UpdateView.clone(TaskMixin),
        TaskCreateView,
        crudlfap.DetailView.clone(
            authenticate=False,
        ),
        crudlfap.ListView.clone(
            authenticate=False,
            filter_fields=['user'],
            search_fields=['user__username', 'about'],
        ),
    ]

    def get_queryset(self, view):
        user = view.request.user

        if user.is_staff or user.is_superuser:
            return self.model.objects.all()
        elif not user.is_authenticated:
            return self.model.objects.none()

        return self.model.objects.filter(user=user)


TaskRouter().register()



from django.utils.translation import gettext_lazy as _

from crudlfap import shortcuts as crudlfap


class Home(crudlfap.TemplateView):
    template_name = 'crudlfap/home.html'
    urlname = 'home'
    urlpath = ''
    title = _('Home')
    title_heading = ''
    material_icon = 'home'
    authenticate = False



from channels.http import AsgiRequest

class Request:
    def __init__(self, asgi_request, client, view):
        self.client = client
        self.view = view
        self.__dict__.update(asgi_request.__dict__)
        self.POST = asgi_request.POST
        self.GET = asgi_request.GET
        self.request = asgi_request



#! /usr/bin/env python

import ast
import inspect
import textwrap
from . import formater

def scope(func):
    func.scope = True
    return func

class JSError(Exception):
    pass

class JS(object):

    name_map = {
        'self'   : 'this',
        'True'   : 'true',
        'False'  : 'false',
        'None'  : 'null',

        'int' : '_int',
        'float' : '_float',

        'super' : '_super',

        # ideally we should check, that this name is available:
        'py_builtins' : '___py_hard_to_collide',
    }

    builtin = set([
        'NotImplementedError',
        'ZeroDivisionError',
        'AssertionError',
        'AttributeError',
        'RuntimeError',
        'ImportError',
        'TypeError',
        'ValueError',
        'NameError',
        'IndexError',
        'KeyError',
        'StopIteration',

        '_int',
        '_float',
        'max',
        'min',
        'sum',
    ])

    bool_op = {
        'And'    : '&&',
        'Or'     : '||',
    }

    unary_op = {
        'Invert' : '~',
        'Not'    : '!',
        'UAdd'   : '+',
        'USub'   : '-',
    }

    binary_op = {
        'Add'    : '+',
        'Sub'    : '-',
        'Mult'   : '*',
        'Div'    : '/',
        'Mod'    : '%',
        'LShift' : '<<',
        'RShift' : '>>',
        'BitOr'  : '|',
        'BitXor' : '^',
        'BitAnd' : '&',
    }

    comparison_op = {
            'Eq'    : "==",
            'NotEq' : "!=",
            'Lt'    : "<",
            'LtE'   : "<=",
            'Gt'    : ">",
            'GtE'   : ">=",
            'Is'    : "===",
            'IsNot' : "is not", # Not implemented yet
        }

    def __init__(self, context=None):
        self.__formater = formater.Formater()
        self.write = self.__formater.write
        self.read = self.__formater.read
        self.indent = self.__formater.indent
        self.dedent = self.__formater.dedent
        self.dummy = 0
        self.classes = ['dict', 'list', 'tuple']
        # This is the name of the class that we are currently in:
        self._class_name = None

        # This lists all variables in the local scope:
        self._scope = []
        #All calls to names within _class_names will be preceded by 'new'
        self._class_names = set()
        self._classes = {}

        if context:
            _copy_context = context.copy()
            for key in _copy_context.keys():
                if callable(_copy_context[key]):
                    value = context.pop(key)
                    source = inspect.getsource(value)
                    source = textwrap.dedent(source)
                    tree = ast.parse(source)
                    sub = JS()
                    sub.visit(tree)
                    self.write(sub.read())

        self._context = context

    def new_dummy(self):
        dummy = "__dummy%d__" % self.dummy
        self.dummy += 1
        return dummy

    def name(self, node):
        return node.__class__.__name__

    def get_bool_op(self, node):
        return self.bool_op[node.op.__class__.__name__]

    def get_unary_op(self, node):
        return self.unary_op[node.op.__class__.__name__]

    def get_binary_op(self, node):
        return self.binary_op[node.op.__class__.__name__]

    def get_comparison_op(self, node):
        return self.comparison_op[node.__class__.__name__]

    def visit(self, node, scope=None):
        try:
            visitor = getattr(self, 'visit_' + self.name(node))
        except AttributeError:
            raise JSError("syntax not supported (%s)" % node)

        if hasattr(visitor, 'statement'):
            return visitor(node, scope)
        else:
            return visitor(node)

    def visit_Module(self, node):
        for stmt in node.body:
            self.visit(stmt)


    @scope
    def visit_FunctionDef(self, node):
        is_static = False
        is_javascript = False
        if node.decorator_list:
            if len(node.decorator_list) == 1 and \
                    isinstance(node.decorator_list[0], ast.Name) and \
                    node.decorator_list[0].id == "JavaScript":
                is_javascript = True # this is our own decorator
            elif self._class_name and \
                    len(node.decorator_list) == 1 and \
                    isinstance(node.decorator_list[0], ast.Name) and \
                    node.decorator_list[0].id == "staticmethod":
                is_static = True
            else:
                raise JSError("decorators are not supported")

        # XXX: disable $def for now, because it doesn't work in IE:
        if self._class_name:
        #if 1:
            if node.args.vararg is not None:
                raise JSError("star arguments are not supported")

            if node.args.kwarg is not None:
                raise JSError("keyword arguments are not supported")

            if node.decorator_list and not is_static and not is_javascript:
                raise JSError("decorators are not supported")

            defaults = [None]*(len(node.args.args) - len(node.args.defaults)) + node.args.defaults

            js_args = []
            js_defaults = []
            self._scope = [arg for arg in node.args.args]

            for arg, default in zip(node.args.args, defaults):
                if not isinstance(arg, ast.arg):
                    raise JSError("tuples in argument list are not supported")

                js_args.append(arg.arg)

                if default is not None:
                    js_defaults.append("%(id)s = typeof(%(id)s) != 'undefined' ? %(id)s : %(def)s;\n" % { 'id': arg.arg, 'def': self.visit(default) })

            if self._class_name:
                prep = "%s.prototype.%s = function(" % \
                        (self._class_name, node.name)
                if not is_static:
                    if not (js_args[0] == "self"):
                        raise NotImplementedError("The first argument must be 'self'.")
                    del js_args[0]
            else:
                prep = "function %s(" % node.name
            self.write(prep + ", ".join(js_args) + ") {")
            self.indent()
            for stmt in js_defaults:
                self.write(stmt)
            for stmt in node.body:
                self.visit(stmt)
            self.dedent()
            self.write('}')

            #If method is static, we also add it directly to the class
            if is_static:
                self.write("%s.%s = %s.prototype.%s;" % \
                        (self._class_name, node.name, self._class_name, node.name))
            #Otherwise, we wrap it to take 'self' into account
            else:
                func_name = node.name
                self.write("%s.%s = function() {" % (self._class_name, func_name))
                self.write("    %s.prototype.%s.apply(arguments[0],Array.slice(arguments,1));"% (self._class_name, func_name))
                self.write("}")

            self._scope = []
        else:
            defaults = [None]*(len(node.args.args) - len(node.args.defaults)) + node.args.defaults

            args = []
            defaults2 = []
            for arg, default in zip(node.args.args, defaults):
                if not isinstance(arg, ast.arg):
                    raise JSError("tuples in argument list are not supported")
                if default:
                    defaults2.append("%s: %s" % (arg.id, self.visit(default)))
                args.append(arg.arg)
            defaults = "{" + ", ".join(defaults2) + "}"
            args = ", ".join(args)
            self.write("var %s = $def(%s, function(%s) {" % (node.name,
                defaults, args))
            self._scope = [arg.arg for arg in node.args.args]
            self.indent()
            for stmt in node.body:
                self.visit(stmt)
            self.dedent()
            self.write("});")

    @scope
    def visit_ClassDef(self, node):
        bases = [self.visit(n) for n in node.bases]
        if not bases:
            bases = ['object']
        class_name = node.name
        #self._classes remembers all classes defined
        self._classes[class_name] = node
        self._class_names.add(class_name)
        self.write("function %s() {" % class_name)
        self.write("    if( this === _global_this){")
        self.write("        t = new %s();" % class_name)
        self.write("        t.__init__.apply(t,arguments);")
        self.write("        return t;")
        self.write("    }")
        self.write("}")
        self.write("%s.__name__ = '%s';" % (class_name, class_name))
        self.write("%s.__setattr__ = object.prototype.__setattr__;" % (class_name))
        self.write("%s.__getattr__ = object.prototype.__setattr__;" % (class_name))
        self.write("%s.__call__    = object.prototype.__call__;" % (class_name))
        self.write("%s.prototype.__class__ = %s;" % (class_name, class_name))
        self.write("%s.prototype.toString = _iter.prototype.toString;" % \
                (class_name))
        from ast import dump
        #~ methods = []
        self._class_name = class_name
        for stmt in node.body:
            if isinstance(stmt, ast.Assign):
                value = self.visit(stmt.value)
                for t in stmt.targets:
                    var = self.visit(t)
                    self.write("%s.%s = %s;" % (class_name, var, value))
                    self.write("%s.prototype.%s = %s.%s;" % (class_name, var, class_name, var))
            else:
                self.visit(stmt)
        self._class_name = None

        #The following is unnecessary: __init__ is inherited from
        #'object'
        #~ methods_names = [m.name for m in methods]
        #~ if not "__init__" in methods_names:
            #~ # if the user didn't define __init__(), we have to add it ourselves
            #~ # because we call it from the constructor above
            #~ self.write("_%s.prototype.__init__ = function() {" % class_name)
            #~ js.append("}")

        self.write('extend(%s, [%s]);' % (class_name, ', '.join(bases)))

    def visit_Return(self, node):
        if node.value is not None:
            self.write("return %s;" % self.visit(node.value))
        else:
            self.write("return;")

    def visit_Delete(self, node):
        return node

    @scope
    def visit_Assign(self, node):
        assert len(node.targets) == 1
        target = node.targets[0]
        #~ if self._class_name:
            #~ target = self._class_name + '.' + target
        value = self.visit(node.value)
        if isinstance(target, (ast.Tuple, ast.List)):
            dummy = self.new_dummy()
            self.write("var %s = %s;" % (dummy, value))

            for i, target in enumerate(target.elts):
                var = self.visit(target)
                declare = ""
                if isinstance(target, ast.Name):
                    if not (var in self._scope):
                        self._scope.append(var)
                        declare = "var "
                self.write("%s%s = %s.__getitem__(%d);" % (declare,
                    var, dummy, i))
        elif isinstance(target, ast.Subscript) and isinstance(target.slice, ast.Index):
            # found index assignment
            self.write("%s.__setitem__(%s, %s);" % (self.visit(target.value),
                self.visit(target.slice), value))
        elif isinstance(target, ast.Subscript) and isinstance(target.slice, ast.Slice):
            # found slice assignmnet
            self.write("%s.__setslice__(%s, %s, %s);" % (self.visit(target.value),
                self.visit(target.slice.lower), self.visit(target.slice.upper),
                value))
        else:
            var = self.visit(target)
            if isinstance(target, ast.Name):
                if not (var in self._scope):
                    self._scope.append(var)
                    declare = "var "
                else:
                    declare = ""
                self.write("%s%s = %s;" % (declare, var, value))
            elif isinstance(target, ast.Attribute):
                js = self.write("%s.__setattr__(\"%s\", %s);" % (self.visit(target.value), str(target.attr), value))
            else:
                raise JSError("Unsupported assignment type")

    def visit_AugAssign(self, node):
        # TODO: Make sure that all the logic in Assign also works in AugAssign
        target = self.visit(node.target)
        value = self.visit(node.value)

        if isinstance(node.op, ast.Pow):
            self.write("%s = Math.pow(%s, %s);" % (target, target, value))
        if isinstance(node.op, ast.FloorDiv):
            self.write("%s = Math.floor((%s)/(%s));" % (target, target, value))

        self.write("%s %s= %s;" % (target, self.get_binary_op(node), value))

    @scope
    def visit_For(self, node):
        if not isinstance(node.target, ast.Name):
            raise JSError("argument decomposition in 'for' loop is not supported")

        for_target = self.visit(node.target)
        for_iter = self.visit(node.iter)

        iter_dummy = self.new_dummy()
        orelse_dummy = self.new_dummy()
        exc_dummy = self.new_dummy()

        self.write("var %s = iter(%s);" % (iter_dummy, for_iter))
        self.write("var %s = false;" % orelse_dummy)
        self.write("while (1) {")
        self.write("    var %s;" % for_target)
        self.write("    try {")
        self.write("        %s = %s.next();" % (for_target, iter_dummy))
        self.write("    } catch (%s) {" % exc_dummy)
        self.write("        if (isinstance(%s, py_builtins.StopIteration)) {" % exc_dummy)
        self.write("            %s = true;" % orelse_dummy)
        self.write("            break;")
        self.write("        } else {")
        self.write("            throw %s;" % exc_dummy)
        self.write("        }")
        self.write("    }")
        self.indent()
        for stmt in node.body:
            self.visit(stmt)
        self.dedent()
        self.write("}")

        if node.orelse:
            self.write("if (%s) {" % orelse_dummy)
            self.indent()
            for stmt in node.orelse:
                self.visit(stmt)
            self.dedent()
            self.write("}")

    @scope
    def visit_While(self, node):

        if not node.orelse:
            self.write("while (%s) {" % self.visit(node.test))
        else:
            orelse_dummy = self.new_dummy()

            self.write("var %s = false;" % orelse_dummy)
            self.write("while (1) {");
            self.write("    if (!(%s)) {" % self.visit(node.test))
            self.write("        %s = true;" % orelse_dummy)
            self.write("        break;")
            self.write("    }")
        self.indent()
        for stmt in node.body:
            self.visit(stmt)
        self.dedent()

        self.write("}")

        if node.orelse:
            self.write("if (%s) {" % orelse_dummy)
            self.indent()
            for stmt in node.orelse:
                self.visit(stmt)
            self.dedent()
            self.write("}")

    @scope
    def visit_If(self, node):
        self.write("if (py_builtins.bool(%s)) {" % self.visit(node.test))
        self.indent()
        for stmt in node.body:
            self.visit(stmt)
        self.dedent()
        if node.orelse:
            self.write("} else {")
            self.indent()
            for stmt in node.orelse:
                self.visit(stmt)
            self.dedent()
        self.write("}")

    @scope
    def _visit_With(self, node):
        pass

    @scope
    def _visit_Raise(self, node):
        pass

    @scope
    def _visit_TryExcept(self, node):
        pass

    @scope
    def _visit_TryFinally(self, node):
        pass

    def visit_Assert(self, node):
        test = self.visit(node.test)

        if node.msg is not None:
            self.write("assert(%s, %s);" % (test, self.visit(node.msg)))
        else:
            self.write("assert(%s);" % test)

    def _visit_Import(self, node):
        pass

    def _visit_ImportFrom(self, node):
        pass

    def _visit_Exec(self, node):
        pass

    def visit_Global(self, node):
        self._scope.extend(node.names)

    def visit_Expr(self, node):
        self.write(self.visit(node.value) + ";")

    def visit_Pass(self, node):
        self.write("/* pass */")

    def visit_Break(self, node):
        self.write("break;")

    def visit_Continue(self, node):
        self.write("continue;")

    def visit_arguments(self, node):
        return ", ".join([self.visit(arg) for arg in node.args])

    def visit_Lambda(self, node):
        return "function(%s) {return %s}" % (self.visit(node.args), self.visit(node.body))

    def visit_BoolOp(self, node):
        return self.get_bool_op(node).join([ "(%s)" % self.visit(val) for val in node.values ])

    def visit_UnaryOp(self, node):
        return "%s(%s)" % (self.get_unary_op(node), self.visit(node.operand))

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Mod) and isinstance(node.left, ast.Str):
            left = self.visit(node.left)
            if isinstance(node.right, (ast.Tuple, ast.List)):
                right = self.visit(node.right)
                return "vsprintf(js(%s), js(%s))" % (left, right)
            else:
                right = self.visit(node.right)
                return "sprintf(js(%s), %s)" % (left, right)
        left = self.visit(node.left)
        right = self.visit(node.right)

        if isinstance(node.op, ast.Pow):
            return "Math.pow(%s, %s)" % (left, right)
        if isinstance(node.op, ast.FloorDiv):
            return "Math.floor((%s)/(%s))" % (left, right)

        return "(%s)%s(%s)" % (left, self.get_binary_op(node), right)

    def visit_Compare(self, node):
        assert len(node.ops) == 1
        assert len(node.comparators) == 1
        op = node.ops[0]
        comp = node.comparators[0]
        if isinstance(op, ast.In):
            return "%s.__contains__(%s)" % (
                    self.visit(comp),
                    self.visit(node.left),
                    )
        elif isinstance(op, ast.NotIn):
            return "!(%s.__contains__(%s))" % (
                    self.visit(comp),
                    self.visit(node.left),
                    )
        elif isinstance(op, ast.Eq):
            return "py_builtins.eq(%s, %s)" % (
                    self.visit(node.left),
                    self.visit(comp),
                    )
        elif isinstance(op, ast.NotEq):
            #In fact, we'll have to override this too:
            return "!(py_builtins.eq(%s, %s))" % (
                    self.visit(node.left),
                    self.visit(comp),
                    )
        else:
            return "%s %s %s" % (self.visit(node.left),
                    self.get_comparison_op(op),
                    self.visit(comp)
                    )

    def visit_Name(self, node):
        id = node.id
        try:
            id = self.name_map[id]
        except KeyError:
            pass

        if id in self.builtin:
            id = "py_builtins." + id;

        #~ if id in self._classes:
            #~ id = '_' + id;
        elif self._context and id in self._context:
            return "str(\"%s\")" % (self._context[id])

        return id

    def visit_Constant(self, node):
        if isinstance(node.value, str):
            return self.visit_Str(node)
        elif isinstance(node.value, bool):
            return 'true' if node.value else 'false'
        else:
            return self.visit_Num(node)

    def visit_Num(self, node):
        return str(node.n)

    def visit_Str(self, node):
        # Uses the Python builtin repr() of a string and the strip string type
        # from it. This is to ensure Javascriptness, even when they use things
        # like b"\\x00" or u"\\u0000".
        return "str(%s)" % repr(node.s).lstrip("urb")

    def visit_Call(self, node):
        func = self.visit(node.func)
        #~ if func in self._class_names:
            #~ func = 'new '+func
        if node.keywords:
            keywords = []
            for kw in node.keywords:
                keywords.append("%s: %s" % (kw.arg, self.visit(kw.value)))
            keywords = "{" + ", ".join(keywords) + "}"
            js_args = ",".join([ self.visit(arg) for arg in node.args ])
            return "%s.args([%s], %s)" % (func, js_args,
                    keywords)
        else:
            if getattr(node, 'starargs', None) is not None:
                raise JSError("star arguments are not supported")

            if getattr(node, 'kwargs', None) is not None:
                raise JSError("keyword arguments are not supported")

            try:
                js_args = ",".join([ self.visit(arg) for arg in node.args ])
            except:
                raise

            return "%s(%s)" % (func, js_args)

    def visit_Raise(self, node):
        assert node.inst is None
        assert node.tback is None
        self.write("throw %s;" % self.visit(node.type))

    def visit_Print(self, node):
        assert node.dest is None
        assert node.nl
        values = [self.visit(v) for v in node.values]
        values = ", ".join(values)
        self.write("py_builtins.print(%s);" % values)

    def visit_Attribute(self, node):
        return "%s.%s" % (self.visit(node.value), node.attr)

    def visit_Tuple(self, node):
        els = [self.visit(e) for e in node.elts]
        return "tuple([%s])" % (", ".join(els))

    def visit_Dict(self, node):
        els = []
        for k, v in zip(node.keys, node.values):
            if isinstance(k, ast.Name):
                els.append('tuple(["%s", %s])' % (self.visit(k), self.visit(v)))
            else:
                els.append("tuple([%s, %s])" % (self.visit(k), self.visit(v)))
        return "dict(tuple([%s]))" % (",\n".join(els))

    def visit_List(self, node):
        els = [self.visit(e) for e in node.elts]
        return "list([%s])" % (", ".join(els))

    def visit_Slice(self, node):
        if node.lower and node.upper and node.step:
            return "slice(%s, %s, %s)" % (self.visit(node.lower),
                    self.visit(node.upper), self.visit(node.step))
        if node.lower and node.upper:
            return "slice(%s, %s)" % (self.visit(node.lower),
                    self.visit(node.upper))
        if node.upper and not node.step:
            return "slice(%s)" % (self.visit(node.upper))
        if node.lower and not node.step:
            return "slice(%s, null)" % (self.visit(node.lower))
        if not node.lower and not node.upper and not node.step:
            return "slice(null)"
        raise NotImplementedError("Slice")

    def visit_Subscript(self, node):
        return "%s.__getitem__(%s)" % (self.visit(node.value), self.visit(node.slice))

    def visit_Index(self, node):
        return self.visit(node.value)

def convert_py2js(s, context=None):
    """
    Takes Python code as a string 's' and converts this to JavaScript.

    Example:

    >>> convert_py2js("x[3:]")
    'x.__getitem__(slice(3, null));'

    """
    t = ast.parse(textwrap.dedent(s))
    v = JS(context)
    v.visit(t)
    return v.read()



from ryzom.components import *
from ryzom.components.django import Static
from ryzom.components.mdc import *
from ryzom.reactive import ReactiveDiv

from .home import Welcome
from .task import TaskList, TaskForm


class Document(Html):
    def __init__(self, view):
        bs_src = 'ryz_ex/css/bootstrap.min.css'
        mdc_js = 'todos/mdc.js'
        mdc_css = 'todos/mdc.css'
        content = [
            Head(
                Title('TODOs'),
                Meta(**{'charset': 'utf-8'}),
                Link(**{
                    'rel': "stylesheet",
                    'href': "https://fonts.googleapis.com/icon?family=Material+Icons",
                }),
                Link(**{'rel': 'stylesheet', 'href': Static(bs_src)}),
                Script(**{
                    'type': 'module',
                    'src': Static(mdc_js)
                }),
                Link(**{
                    'rel': 'stylesheet',
                    'href': Static(mdc_css)
                }),
            ),
            # Body(Base(view))
            Body(
                MdcTopAppBar(
                    'Ryzom Todos',
                    action_items=[
                        ('search', 'Search', 'url', 'events',),
                        ('more_vert', 'Options', None, [
                                ('favorite', 'Favourite', 'url', 'events',),
                                ('favorite', 'Favourite', 'url', 'events',),
                            ],
                        ),
                    ],
                ),
                MdcDrawer(
                    MdcDrawerHeader(
                        drawer_title="Title",
                        drawer_subtitle="subtitle",
                    ),
                    MdcNavList(
                        MdcListItem(
                            Text("Home"),
                            icon='home',
                            href='/',
                            active=True,
                        ),
                        MdcListItem(
                            Text("Todos"),
                            icon='task',
                            href='/todos',
                        ),
                    )
                ),
                MdcAppContent(
                    Base(view),
                ),
            )
        ]
        super().__init__(*content)


class Base(Div):
    def __init__(self, view):
        attrs = {
            'class': 'container',
            '_id': 'tasklist_container',
        }
        content = [
            A(Nav(H1('Todos'),
                  **{'class': 'navbar navbar-light bg-light'}),
              **{'href': '/'}
              ),
            ReactiveDiv('main-container', view, TaskList(), TaskForm()
                        ) if view.url == 'todos'
            else ReactiveDiv('main-container', view, Welcome())
        ]
        super().__init__(*content, **attrs)



from django.apps import apps
from django.contrib.auth.models import User

from ryzom.methods import Methods

# Avoid importing models during django_setup().
AppModel = apps.get_model('todos.Task')


def create_user(user, params):
    username = params.get('username', None)
    password = params.get('username', None)
    email = params.get('username', None)

    if username and password and email:
        user, created = User.objects.get_or_create(
                username=username, email=email)
        if created:
            user.set_password(password)
            user.save()
            return user.id
        else:
            return {
                'type': 'Error',
                'params': {
                    'name': 'Forbidden',
                    'message': 'User already exists in Database'
                }
            }
    else:
        return {
            'type': 'Error',
            'params': {
                'name': 'Bad request',
                'message': 'You must provide a unique username, \
                        a valid email address and a password'
            }
        }


def insert_task(user, params):
    if 'about' in params:
        user = user if user.id else None
        t = AppModel(user=user, about=params['about'])
        t.save()
        return True
    return False


def remove_task(user, params):
    if 'id' in params:
        t = AppModel.objects.get(id=params['id'])
        if t:
            t.delete()
            return True
    return False


Methods.add({
    'create_user': create_user,
    'insert_task': insert_task,
    'remove_task': remove_task
})



'''
Defines the ryzom View class and the main index view
'''
from django import http
from django.contrib.staticfiles import finders

from ryzom.components import Component, Script
from ryzom.models import Subscription, Publication, Clients

ryzom_js_file = finders.find('ryzom/js/ryzom.js')
ryzom_js_fd = open(ryzom_js_file, 'r')
ryzom_js = f'{ryzom_js_fd.read()}\n'

py2js_js_file = finders.find('ryzom/js/py-builtins.js')
py2js_js_fd = open(py2js_js_file, 'r')
py2js_js = f'{py2js_js_fd.read()}\n'


class View:
    '''
    View class, inherit from this class to create a ryzom View.
    the only parameter it take is the channel name of the client
    creating the view instance.
    It is necessary to overload the methods onurl and render when
    inheriting from this class.
    The methods oncreate() and ondestroy() can also be overloaded
    to run code on those events.

    :param str channel_name: The name of the channel this instance \
            is attached to
    '''
    def __init__(self, request, *args, **kwargs):
        self.request = request
        self.args = args
        self.kwargs = kwargs
        self.scripts = ''

    def onurl(self, url):
        '''
        To be overloaded.
        This method will be called whenever a websocket url is required
        (see ryzom.consumers.recv_geturl) and routed to the same view
        controller.
        Its aim is to update the instance inheriting from this class
        on an geturl call.

        :param str url: The url the client is trying to access
        '''
        self.url = url
        return True

    def render(self, request):
        '''
        To be overloaded.
        This method will be called whenever a websocket url is required
        and router to a new View controller.
        Its aim is to render a full content. After that, only reactive
        content should be updated by the onurl() method.

        :param str url: The url the client is trying to access
        '''
        raise NotImplementedError

    def renderHTML(self, url, qs):
        content = self.render(self.request)
        if isinstance(content, Component):
            return http.HttpResponse(
                '<!DOCTYPE html>' +
                content.to_html(self) +
                self.renderJS()
            )

        return content

    def renderJS(self):
        return (
            '<script type="text/javascript">'
            + ryzom_js
            + py2js_js
            + self.scripts
            + '</script>'
        )

    def oncreate(self, url):
        '''
        Hook to optionaly overload.
        This method will be called whenever a new instance of a View is
        created by the consumer.
        Its aim is to setup the view data, such as content, url or whatever.

        :param str url: The url the client is trying to access
        '''
        self.url = url

    def ondestroy(self):
        '''
        Hook to optionaly overload.
        This method will be called whenever a instance of a View is
        destroyed by the consumer.
        '''
        pass

    @classmethod
    def as_view(cls, request, *args, **kwargs):
        view = cls(request, *args, **kwargs)
        view.oncreate(request.path_info)
        view.onurl(request.path_info)

        return view.renderHTML(request.path_info, request.GET.urlencode())


class ReactiveView(View):
    def __init__(self, request):
        self.reactive_components = {}
        super().__init__(request)

    def addReactiveComponent(self, component):
        '''
        Method intended to be used by reactive components.
        Registers a component to the reactive_components dict
        of the view instance.

        :param ReactiveComponent component: The reactive component to \
                attach to this View instance
        '''
        self.reactive_components[component.name] = component

    def reactive(self, name, content):
        '''
        Method to update the content of a reactive component by name.
        the content parameter must be a subclass of ryzom.components.Component

        :param str name: The name of the \
                :class:`ryzom.reactive.ReactiveComponent` to update
        :param Component content: The new content
        '''
        component = self.reactive_components[name]
        component.setcontent(content)

    def renderHTML(self, url, qs):
        def find_tag(c, t):
            if c.tag == t:
                return c
            for children in c.content:
                if c.tag != 'text' and not isinstance(children, str):
                    tag = find_tag(children, t)
                    if tag:
                        return tag

        content = self.render(self.request)

        script = '\n'
        script += f'token = "{self.request.client.token}";\n'
        script += f'current_url = "{url}";\n'
        script += f'query_string = "{qs}";\n'
        script += ryzom_js
        script += self.renderScript(content)

        body = find_tag(content, 'body')
        body.addchild(Script(script))
        return content.to_html(self)

    def renderScript(self, content, set_tag=True):
        script = '(function() {' if set_tag else ''
        content = content if isinstance(content, list) else [content]
        for component in content:
            if isinstance(component, str):
                continue
            if component.publication:
                component.create_subscription(self)
                pub = component.publication
                sub = component.subscription
                script += f'ryzom.subscribe("{pub}", "{sub.id}", "{component._id}",'
                script += 'function(r,e){if (e) {console.log(e);}});\n'
            for event, cb in component.events.items():
                script += f'getElementByUuid("{component._id}").'
                script += f'addEventListener("{event}",' + \
                          f'function(e){{{cb}}});\n'

            if component.tag != 'text':
                script += self.renderScript(component.content, False)
        if set_tag:
            script += '})();'
        return script


def index(request, url=''):
    '''
    Default HTTP django view.
    This view renders the main ryzom template (index.html) to the client.
    index.html contains all the core JS ryzom logic in a basic HTML document.
    it defines the 'url' and 'query_string' variable, templated in the document
    to allow the JS router to know where it is.
    '''
    if request.ryzom.client is None:
        request.ryzom.client = Clients.objects.create()

    view = request.ryzom.view(request.ryzom)
    view.oncreate(url)
    view.onurl(url)

    return view.renderHTML(url, request.GET.urlencode())



'''
Ryzom components declarations.
There's still a lot of tags missing.
They will be added when they'll be needed
'''
import copy
import importlib
import re
import uuid


def component_html(path, *args, **kwargs):
    from django.utils.safestring import mark_safe
    try:
        from jinja2.utils import Markup
    except ImportError:
        Markup = None
    import cli2
    ComponentCls = cli2.Node.factory(path).target
    component = ComponentCls(*args, **kwargs)
    html = component.to_html()

    if Markup:
        html = Markup(html)
    return mark_safe(html)


class HTMLPayload(dict):
    def __init__(self, **kwargs):
        super().__init__(**{
            k.replace('_', '-'): v for k, v in kwargs.items()
        })

    def __getattr__(self, name):
        html_name = name.replace('_', '-')
        if html_name in self:
            return self[html_name]
        raise AttributeError(f'{self} object has no attribute {name}')

    def __setattr__(self, name, value):
        self.__setitem__(name, value)

    def __setitem__(self, name, value):
        name = name.replace('_', '-')
        if value is None:
            if name in self:
                del self[name]
        else:
            super().__setitem__(name, value)

    def update(self, other):
        for key, value in other.items():
            self[key] = value


class CStyle(HTMLPayload):
    @classmethod
    def to_dict(cls, value):
        result = {}
        for rule in value.split(';'):
            if not rule.strip():
                continue
            key, value = rule.split(':')
            result[key.strip()] = value.strip()
        return result


class CAttrs(HTMLPayload):
    def __getitem__(self, name):
        if name == 'style' and 'style' not in self:
            # Create CStyle on the fly
            self['style'] = CStyle()

        if name == 'cls':
            # Translate 'cls' into class
            name = 'class'

        return super().__getitem__(name)

    def __getattr__(self, name):
        if name == 'style' and 'style' not in self:
            self['style'] = CStyle()

        if name == 'cls':
            name = 'class'

        return super().__getattr__(name)

    def __setitem__(self, name, value):
        if name == 'cls':
            # Maintain the "class" attribute
            name = 'class'
        elif name == 'addcls':
            if 'class' not in self:
                self['class'] = value
            elif self['class']:
                self['class'] += ' ' + value
            else:
                self['class'] = value
            return
        elif name == 'rmcls':
            if 'class' in self:
                self['class'] = self['class'].replace(
                    value, ''
                ).replace('  ', ' ').strip()
            return

        if name == 'style' and isinstance(value, str):
            self.style = CStyle.to_dict(value)
        else:
            super().__setitem__(name, value)

    def update(self, other):
        for key, value in other.items():
            if key == 'style':
                if isinstance(value, str):
                    value = CStyle.to_dict(value)
                self['style'].update(value)
            else:
                self[key] = value

    def to_html(self):
        result = []
        for key, value in self.items():
            if key == 'style':
                new_value = []
                for style_key, style_value in value.items():
                    new_value.append(f'{style_key}: {style_value}')
                value = '; '.join(new_value)

            if value is True:
                result.append(key)
            elif value is not False:
                result.append(f'{key}="{value}"')
        return ' '.join(result)


class ComponentMetaclass(type):
    def __new__(cls, name, bases, class_attrs):
        attrs = CAttrs()
        for base in bases:
            if base_attrs := getattr(base, 'attrs', None):
                attrs.update(base_attrs)
        if class_attrs.get('attrs', None):
            attrs.update(class_attrs['attrs'])

        if 'style' in class_attrs:
            attrs.update(dict(style=class_attrs['style']))

        class_attrs['attrs'] = attrs

        return super().__new__(cls, name, bases, class_attrs)


class Component(metaclass=ComponentMetaclass):
    '''Main ryzom component 'abstract' class to be inherited.

    This class defines the common attrsibutes and methods to all
    components,the main one being to_obj() that format an instance
    as a serializable dict that can be sent to the client over websocket
    the to_html() method is missing for now, it will be useful when
    implementing Server-Side Rendering

    Upon creation, if not specified, an instance of the Component class
    is considered as a child of the <html> tag, this is not guaranteed
    to be kept in a near future, because it's totally useless.
    Being a childnode of <body> seem a lot more meaningful.
    If no _id is specified, a random (but still unique) one will be
    generated.

    :param str tag: The HTML tag of the component
    :param list<Component> content: The component instances contained \
            by the current instance
    :param dict attrs: HTML attrsibutes (id, class, style, ...)
    :param dict events: The events to add listeners to \
            (click, hover, ...)
    :param str parent: The id of the component that contains the \
            current instance
    :param str _id: The _id of the current instance (must be unique)
    '''

    __publication = None

    def __getattr__(self, name):
        '''Instanciate attrs and/or style on the fly.'''
        if name == 'attrs':
            self.attrs = CAttrs()
            return self.attrs
        elif name == 'style':
            self.style = self.attrs.style = CStyle()
            return self.style
        return super().__getattr__(name)

    def __init__(self, *content, **attrs):
        cls = type(self)
        self.content = list(content) or []

        self._id = attrs.pop('_id', uuid.uuid1().hex)
        self.parent = attrs.pop('parent', None)

        self.__dict__['tag'] = getattr(self, 'tag', None)
        if 'tag' in attrs:
            self.tag = attrs.pop('tag')
        elif not getattr(self, 'tag', None):
            s1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', type(self).__name__)
            self.tag = re.sub('([a-z0-9])([A-Z])', r'\1-\2', s1).lower()

        self.__dict__['selfclose'] = attrs.pop(
            'selfclose',
            getattr(self, 'selfclose', False),
        )

        self.noclose = self.tag.lower() in [
            'area', 'base', 'br', 'col', 'embed', 'hr', 'img',
            'input', 'link', 'meta', 'param', 'source', 'track', 'wbr',
        ]

        self.events = attrs.pop('events', {})

        # create an instance attribute from the class attribute
        self.__dict__['attrs'] = copy.deepcopy(getattr(self, 'attrs', {}))

        # update with kwargs
        self.attrs.update(attrs)

        self.position = 0

        self.preparecontent()

    def preparecontent(self):
        '''Set the parent and position of children

        meant to be called by __init__().
        This method sets the current component as parent of each child
        Moreover it sets the child's position attribute to its index
        in the current component's content list
        '''
        # handle text node as content
        if self.tag == 'text':
            self.content = self.content[0]
        else:
            for i, c in enumerate(self.content):
                if isinstance(c, str):
                    self.content[i] = c = Text(c)
                c.parent = self._id
                c.position = i

    def addchild(self, component):
        '''Add a child component

        Currently not used at all, but meant to push a new child
        at the end of the content's list
        A call to preparecontent() should follow the call to this method

        :param Component component: The child component to add to the \
                content of the current instance
        '''
        component.position = len(self.content)
        component.parent = self._id
        self.content.append(component)

    def addchildren(self, components):
        '''Add a list of children

        Currently not used at all, meant to push children at the
        end of the content's list.
        A call to preparecontent() should follow the call to this method

        :param list(Component) components: The component list to insert \
                in the content of the current instance
        '''
        for component in components:
            self.addchild(component)

    def addevents(self, events):
        '''Add events to instance

        Currently not used, meant to add/update a dict of eventListener
        attached to the DOM element associated with this instance

        :param dict events: The dict to update events with
        '''
        self.events.update(events)

    def to_obj(self, context=None):
        '''Get a serializable dict of the instance

        This methods returns a dict representation of the current
        instance. I handles subscriptions that will have this component
        instance _id as parent attribute.
        Recursively sets the content as dict too (maybe recursiveness is not
        a good thing to do without any control of how deep can the tree be,
        there's a risk of stack overflow that we must keep in mind)

        :returns: A serializable representation of the instance
        '''
        sub = None
        if self.publication:
            from ..models import Subscriber, Subscription, Publication
            pub = Publication.objects.get(name=self.publication)
            Subscriber.objects.get_or_create(
                parent_id=self._id,
                parent_module=self.__module__,
                parent_class=self.__class__.__name__
            )
            model_mod = importlib.import_module(pub.model_module)
            model = getattr(model_mod, pub.model_class)
            func = getattr(model, pub.name)
            qs = self.subscribe(None, func(), None)
            self.content = []
            tmpl_module = importlib.import_module(pub.template_module)
            tmpl_class = getattr(tmpl_module, pub.template_class)
            for c in qs:
                self.content.append(tmpl_class(c))

            from django.contrib.postgres.aggregates import ArrayAgg
            sub = Subscription.objects.create(
                parent=self._id,
                publication=pub,
                queryset=qs.aggregate(ids=ArrayAgg('id'))['ids'],
                options=None,
                client=None
            )

        return {
            '_id': self._id,
            'tag': self.tag,
            'content': [
                c if isinstance(c, str) else c.to_obj()
                for c in self.content
            ] if self.tag != 'text' else self.content,
            'parent': self.parent,
            'position': self.position,
            'events': self.events,
            'attrs': self.attrs,
            'publication': self.publication,
            'subscription': f"{sub.id}" if sub else None
        }

    @property
    def publication(self):
        return self.__publication

    @publication.setter
    def publication(self, value=None):
        self.__publication = value

    def to_html(self, **kwargs):
        if self.tag == 'text':
            return f'{self.content}'
        attrs = ' '.join([self.attrs.to_html(), f'ryzom-id="{self._id}"'])
        html = ''
        if getattr(self, 'selfclose', False):
            html = f'<{self.tag} {attrs}/>'
        elif getattr(self, 'noclose', False):
            html = f'<{self.tag} {attrs}>'
        else:
            html = f'<{self.tag} {attrs}>'
            if self.publication:
                self.create_subscription(kwargs)
            for c in self.content:
                html += (
                    c.to_html(**kwargs)
                    if getattr(c, 'to_html', None) else str(c)
                )
            html += f'</{self.tag}>'

        js_str = str(self.render_js())
        if js_str:
            js_str = js_str[0:-2]
            js_str += '();'
            #context.scripts += js_str

        return html

    def create_subscription(self, context=None):
        from ..models import Subscriber, Subscription, Publication
        pub = Publication.objects.get(name=self.publication)
        if getattr(self, 'subscription', None):
            return
        else:
            sub = Subscription.objects.filter(
                client=context.request.client,
                publication=pub,
                parent=self._id
            ).first()
            if sub:
                self.subscription = sub
                return
        Subscriber.objects.get_or_create(
            parent_id=self._id,
            parent_module=self.__module__,
            parent_class=self.__class__.__name__
        )
        model_mod = importlib.import_module(pub.model_module)
        model = getattr(model_mod, pub.model_class)
        func = getattr(model, pub.name)
        qs = self.subscribe(None, func(), None)
        self.content = []
        tmpl_module = importlib.import_module(pub.template_module)
        tmpl_class = getattr(tmpl_module, pub.template_class)
        for c in qs:
            self.content.append(tmpl_class(c))

        from django.contrib.postgres.aggregates import ArrayAgg
        self.subscription = Subscription.objects.create(
            parent=self._id,
            publication=pub,
            queryset=qs.aggregate(ids=ArrayAgg('id'))['ids'],
            options=None,
            client=context.request.client
        )

    def render(self, **kwargs):
        return self.to_html(**kwargs)

    def render_js(self):
        return ''

    def render_js_tree(self, lvl=0):
        js_str = str(self.render_js())

        if hasattr(self, 'content'):
            for c in self.content:
                if isinstance(c, Component):
                    js_str += c.render_js_tree(lvl+1)

        if js_str and not lvl:
            js_str = js_str[0:-2]
            js_str += '();'

        return js_str


class CTree(Component):
    def __init__(self, *components):
        self.components = components
        self.__name__ = components[-1].__name__

    def __call__(self, **kwargs):
        print(self.components[-1])
        component = self.components[-1](**kwargs)
        for wrapper in reversed(self.components[:-1]):
            component = wrapper(component, **kwargs)
        return component


class CList(Component):
    def to_html(self, **kwargs):
        html_str = ''
        for c in self.content:
            html_str += c.to_html(**kwargs)

        js_str = str(self.render_js())
        if js_str:
            js_str = js_str[0:-2]
            js_str += '();'
            context.scripts += js_str

        return html_str

    def to_obj(self, context=None):
        content = [
            c.to_obj(context)
            for c in self.content
        ]
        return content


class Text(Component):
    '''
    Text component

    Represents a text node
    '''
    tag = 'text'



from django.conf import settings
from django.db import models

from ryzom.pubsub import Publishable, publish


class Task(Publishable, models.Model):
    user = models.ForeignKey(
        getattr(settings, 'AUTH_USER_MODEL', 'auth.User'),
        on_delete=models.SET_NULL, blank=True, null=True)
    about = models.CharField(max_length=1024)

    class Project:  # app_label - for app-specific model options
        ryzom = True

    def __str__(self):
        return self.about

    @publish('todos.components.task.Task')
    def all_tasks(cls):
        return cls.objects.all()

    @publish('todos.components.task.Task')
    def first_tasks(cls):
        return cls.objects.all().order_by('about')[0:5]


Task.all_tasks()
Task.first_tasks()



from django.middleware.csrf import get_token
from ryzom.html import *
from ryzom.js.renderer import JS


class MDCLink(A):
    pass


class MDCIcon(Icon):
    attrs = {
        'class': 'material-icons',
        'aria-hiddem': 'true'
    }


class MDCTextButton(Button):
    def __init__(self, text, icon=None, **kwargs):
        content = [Span(cls='mdc-button__ripple')]
        if icon:
            content.append(MDCIcon(icon))
        content.append(Span(text, cls='mdc-button__label'))
        super().__init__(
            *content,
            cls='mdc-button',
            **kwargs,
        )


class MDCButtonOutlined(Button):
    def __init__(self, text, p=True, icon=None, **kwargs):
        black = 'black-button' if p else ''
        content = [Span(cls='mdc-button__ripple')]
        if icon and isinstance(icon, str):
            content.append(MDCIcon(icon))
        elif icon:
            content.append(icon)
        content.append(Span(text, cls='mdc-button__label'))
        super().__init__(
            *content,
            cls=f'mdc-button mdc-button--outlined {black}',
            **kwargs
        )


class MDCButton(Button):
    def __init__(self, text, p=True, icon=None, **kwargs):
        black = 'black-button' if p else ''

        if icon and isinstance(icon, str):
            content = [MDCIcon(icon)]
        elif icon:
            content = [icon]
        else:
            content = []

        content.append(Span(text, cls='mdc-button__label'))
        super().__init__(
            *content,
            cls=f'mdc-button mdc-button--raised {black}',
            **kwargs
        )


class MDCButtonLabelOutlined(Label):
    def __init__(self, text, p=True, icon=None):
        black = 'black-button' if p else ''
        content = [Span(cls='mdc-button__ripple')]
        if icon:
            content.append(MDCIcon(icon))
        content.append(Span(text, cls='mdc-button__label'))
        super().__init__(
            *content,
            cls=f'mdc-button mdc-button--outlined {black}'
        )


class MDCTextInput(Input):
    def __init__(self, name, **attrs):
        attrs['name'] = name
        attrs.setdefault('type', 'text')
        attrs.setdefault('class', 'mdc-text-field__input')
        super().__init__(**attrs)


class MDCLabel(Span):
    attrs = {'class': 'mdc-floating-label'}


class MDCTextRipple(Span):
    def __init__(self):
        super().__init__(**{'class': 'mdc-text-field__ripple'})


class MDCLineRipple(Span):
    def __init__(self):
        super().__init__(**{'class': 'mdc-line-ripple'})


class MDCTextFieldFilled(Label):
    attrs = {
        'class': 'mdc-text-field mdc-text-field--filled',
    }

    def __init__(self, label='', input_id='', label_id='', **kwargs):
        content = [
            MDCTextRipple(),
            MDCLabel(label, label_id),
            MDCTextInput(input_id, **kwargs),
            MDCLineRipple(),
        ]
        super().__init__(*content, **self.attrs)


class MDCNotchOutline(Span):
    attrs = {'class': 'mdc-notched-outline'}

    def __init__(self, *content, **attrs):
        super().__init__(
            Span(cls='mdc-notched-outline__leading'),
            Span(
                *content,
                cls='mdc-notched-outline__notch',
            ),
            Span(cls='mdc-notched-outline__trailing'),
            **attrs,
        )


class MDCField(Div):
    VERTICAL_MARGIN = '32px'
    style = f'margin-top: {VERTICAL_MARGIN}; margin_bottom: {VERTICAL_MARGIN}'

    def __init__(self, *content, name, label=None, help_text=None, value=None,
                 errors=None, **attrs):
        helper_id = f'id_{name}_helper'
        errors_id = f'id_{name}_errors'

        if errors:
            self.errors = MDCErrorList(*errors, id=errors_id)
        else:
            self.errors = ''

        if help_text:
            self.help_text = MDCHelpText(help_text, id=helper_id)
        else:
            self.help_text = ''

        super().__init__(*content, self.errors, self.help_text, **attrs)


class MDCTextFieldOutlined(MDCField):
    def __init__(self, html_input, label=None, help_text=None, errors=None, **attrs):
        self.html_input = html_input
        self.html_input.attrs.addcls = 'mdc-text-field__input'

        name = self.html_input.attrs.name
        input_id = f'id_{name}'
        label_id = f'id_{name}_label'
        helper_id = f'id_{name}_helper'
        errors_id = f'id_{name}_errors'

        floating_label = Span(label, id=label_id, cls='mdc-floating-label')
        notch_outline = MDCNotchOutline(floating_label)
        self.label = Label(
            notch_outline,
            self.html_input,
            id=label_id,
            cls='mdc-text-field mdc-text-field--outlined',
            data_mdc_auto_init='MDCTextField',
        )
        self.html_input.attrs.aria_labelledby = label_id

        value = self.html_input.attrs.get('value', '')
        if value not in ('', None):
            # float label because there is a value
            self.label.attrs.addcls = 'mdc-text-field--label-floating'
            floating_label.attrs.addcls = 'mdc-floating-label--float-above'
            notch_outline.attrs.addcls = 'mdc-notched-outline--notched'

        if errors:
            self.label.attrs.addcls = 'mdc-text-field--invalid'
            self.html_input.attrs.aria_describedby = errors_id
            self.html_input.attrs.aria_controls = errors_id
        elif help_text:
            self.html_input.attrs.aria_describedby = helper_id
            self.html_input.attrs.aria_controls = helper_id

        super().__init__(self.label, name=name,
                         help_text=help_text,
                         errors=errors, **attrs)


class MDCTextareaFieldOutlined(MDCTextFieldOutlined):
    def __init__(self, textarea, label=None, help_text=None, errors=None):
        super().__init__(
            textarea,
            label=label,
            help_text=help_text,
            errors=errors,
        )
        # decorate textarea with a resizer
        self.label.content[1] = Span(textarea, cls='mdc-text-field__resizer')
        self.label.attrs.addcls = 'mdc-text-field--textarea'


class MDCFormField(Div):
    attrs = {'class': 'mdc-form-field'}

    def __init__(self, *content, **kwargs):
        super().__init__(*content, **self.attrs, **kwargs)


class MDCFileInput(Div):
    def __init__(self, btn_text='', input_id='', name=''):
        self.btn = MDCButtonLabelOutlined(btn_text, False)
        self.input_id = input_id
        self.btn.attrs['for'] = input_id
        self.selected_text = Span('No file selected')
        super().__init__(
            Span(
                Input(type='file', id=input_id, name=name),
                style='display:block;width:0;height:0;overflow:hidden'
            ),
            self.selected_text,
            self.btn
        )

    def render_js(self):
        def change_event():
            def update_name(event):
                file_name = document.querySelector(input_id).value
                label = getElementByUuid(label_id)

                if file_name != '':
                    setattr(label, 'innerText', file_name)
                else:
                    setattr(label, 'innerText', 'No file selected')


            document.querySelector(input_id).addEventListener('change', update_name)

        return JS(change_event, dict(
            input_id=f'#{self.input_id}',
            label_id=self.selected_text._id,
        ))


class CSRFInput(Input):
    def __init__(self, request):
        super().__init__(
            type='hidden',
            name='csrfmiddlewaretoken',
            value=get_token(request)
        )


class MDCSplitDateTime(Div):
    def __init__(self, label='', input_id='', label_id='', **kwargs):
        name = kwargs.pop('name')
        kwargs.pop('type', None)
        value = kwargs.pop('value', None)
        date = value.date() if value else ''
        time = value.time() if value else ''

        super().__init__(
            MDCTextFieldOutlined(
                label='Date',
                input_id='date_' + name + '_input',
                name=name + '_0',
                value=date,
                type='date',
                **kwargs),
            MDCTextFieldOutlined(
                label='Time',
                input_id='time_' + name + '_input',
                name=name + '_1',
                value=time,
                type='time',
                **kwargs),
        )

    def set_error(self, error):
        helper = MDCTextFieldHelperLine(error, 'alert')
        label = self.content[1]
        label.attrs['class'] += ' mdc-text-field--invalid'
        label.attrs['aria-describedby'] = helper._id
        label.attrs['aria-controls'] = helper._id

        self.content.append(helper)


class MDCTextFieldHelperText(Div):
    attrs = {
        'class': ' '.join([
            'mdc-text-field-helper-text',
            'mdc-text-field-helper-text--persitent',
        ]),
    }

class MDCTextFieldHelperLine(Div):
    attrs = {
        'aria-hidden': 'true',
        'cls': 'mdc-text-field-helper-line',
    }


class MDCVerticalMargin(Div):
    style = 'margin-top: 18px; margin-bottom: 18px'


class MDCList(Div):
    attrs = {'class': 'mdc-list', 'data-mdc-auto-init': 'MDCList'}


class MDCListItem(Li):
    attrs = {'class': 'mdc-list-item'}

    def __init__(self, *content, **attrs):
        super().__init__(
            Span(cls='mdc-list-item__ripple'),
            Span(*content, cls='mdc-list-item__text'),
            **attrs,
        )


class MDCSnackBar(Div):
    def __init__(self, msg, status='success'):
        super().__init__(
            Div(
                Div(
                    msg,
                    cls='mdc-snackbar__label',
                    **{'aria-atomic': 'false'}
                ),
                Div(
                    Button(
                        Div(cls='mdc-button__ripple'),
                        Span('OK', cls='mdc-button__label'),
                        type='button',
                        cls='mdc-button mdc-snackbar__action'
                    ),
                    cls='mdc-snackbar__actions',
                    **{'aria-atomic': 'true'}
                ),
                cls='mdc-snackbar__surface',
                role='status',
                **{'aria-relevant': 'addition'}
            ),
            cls='mdc-snackbar',
            **{'data-mdc-auto-init': 'MDCSnackbar'}
        )

    def render_js(self):
        return (
            '\nvar snack = function() {' +
            '\n\tvar elem = document.querySelector(".mdc-snackbar");' +
            '\n\tsn = new mdc.snackbar.MDCSnackbar(elem);' +
            '\n\tsn.open();' +
            '\n}; '
        )


class MDCErrorListItem(Li):
    style = dict(list_style_type='none')


class MDCErrorList(Ul):
    style = dict(
        padding_left='10px',
        margin_top=0,
        margin_bottom=0,
        font_size='var(--mdc-typography-caption-font-size, 0.75rem)',
        color='var(--mdc-theme-error, #b00020)',
    )
    def __init__(self, *content, **attrs):
        super().__init__(*[
            MDCErrorListItem(c) if isinstance(c, str) else c
            for c in content
        ], **attrs)


class MDCHelpText(Div):
    style = dict(
        font_size='var(--mdc-typography-caption-font-size, 0.75rem)',
        color='rgba(0, 0, 0, 0.6)',
        padding_left='10px',
    )


class MDCCheckboxInput(Div):
    """The actual input HTML element (widget)."""
    def __init__(self, **kwargs):
        kwargs.setdefault('type', 'checkbox')
        super().__init__(
            Input(
                cls='mdc-checkbox__native-control',
                **kwargs
            ),
            Div(
                Component(
                    Component(
                        tag='path', fill='none',
                        d="M1.73,12.91 8.1,19.28 22.79,4.59",
                        cls='mdc-checkbox__checkmark-path'),
                    tag='svg', viewBox='0 0 24 24',
                    cls='mdc-checkbox__checkmark'
                ),
                Div(cls='mdc-checkbox__mixedmark'),
                cls='mdc-checkbox__background',
            ),
            cls='mdc-checkbox',
        )


class MDCCheckboxField(MDCField):
    def __init__(self, *content, name, label=None, help_text=None, value=None,
                 errors=None):
        super().__init__(
            MDCFormField(
                *content,
                Label(label or name.capitalize()),
            ),
            name=name,
            errors=errors,
            help_text=help_text,
        )

        # compensate for widget margin
        if self.errors:
            self.errors.attrs.style.margin_top = '-10px'
        elif self.help_text:
            self.help_text.attrs.style.margin_top = '-10px'


class MDCCheckboxSelectField(MDCField):
    def __init__(self, *content, name, label=None, help_text=None, value=None,
                 errors=None):
        super().__init__(
            *content,
            name=name,
            errors=errors,
            help_text=help_text,
        )

        # compensate for widget margin
        if self.errors:
            self.errors.attrs.style.margin_top = '-10px'
        elif self.help_text:
            self.help_text.attrs.style.margin_top = '-10px'


class MDCCheckboxListItem(Li):
    def __init__(self, title, id, checked=False, **kwargs):
        self.input_id = id
        if checked:
            kwargs['checked'] = True
        super().__init__(
            Span(cls='mdc-list-item__ripple'),
            Span(
                Div(
                    Input(
                        id=id,
                        type='checkbox',
                        cls='mdc-checkbox__native-control',
                        **kwargs),
                    Div(
                        Component(
                            Component(
                                tag='path', fill='none',
                                d="M1.73,12.91 8.1,19.28 22.79,4.59",
                                cls='mdc-checkbox__checkmark-path'),
                            tag='svg', viewBox='0 0 24 24',
                            cls='mdc-checkbox__checkmark'),
                        Div(cls='mdc-checkbox__mixedmark'),
                        cls='mdc-checkbox__background'),
                    cls='mdc-checkbox'),
                cls='mdc-list-item__graphic'),
            Label(title, cls='mdc-list-item__text', **{'for': id}),
            cls='mdc-list-item',
            role='checkbox',
            **{
                'aria-checked': 'true' if checked else 'false',
                'data-list-item-of': id
            }
        )

    def render_js(self):
        def events():
            def click_input(event):
                event.stopPropagation()
                elem = event.target.querySelector('input')
                if elem:
                    elem.click()

            elem = getElementByUuid(id)
            setattr(elem, 'onclick', click_input)

        return JS(events, dict(id=self._id))


class MDCMultipleChoicesCheckbox(Ul):
    def __init__(self, name, choices, n=1, **kwargs):
        self.max = n
        alabel = kwargs.pop('aria-label', 'Label')
        super().__init__(
            *(MDCCheckboxListItem(
                title,
                f'{name}_input_{i}',
                name=name,
                value=value,
                **kwargs
            ) for i, title, value in choices),
            cls='mdc-list',
            role='group',
            **{'aria-label': alabel}
        )

    def render_js(self):
        def change_event():
            def update_inputs(event):
                checked = input_list.querySelectorAll('input:checked')
                unchecked = input_list.querySelectorAll('input:not(:checked)')

                def disable(elem, pos, arr):
                    setattr(elem, 'disabled', 'true')
                    list_item = document.querySelector(
                        '[data-list-item-of="' + elem.id + '"]'
                    ).classList.add('mdc-list-item--disabled')

                def enable(elem, pas, arr):
                    setattr(elem, 'disabled', undefined)
                    list_item = document.querySelector(
                        '[data-list-item-of="' + elem.id + '"]'
                    ).classList.remove('mdc-list-item--disabled')

                if checked.length >= max_choices:
                    unchecked.forEach(disable)
                else:
                    unchecked.forEach(enable)

            input_list = getElementByUuid(id)
            input_list.addEventListener('change', update_inputs)

            document.addEventListener('readystatechange', update_inputs)

        return JS(change_event, dict(id=self._id, max_choices=self.max))


class MdcTopAppBar(Component):
    """
    <header class="mdc-top-app-bar">
      <div class="mdc-top-app-bar__row">
        <section class="mdc-top-app-bar__section mdc-top-app-bar__section--align-start">
          <button class="material-icons mdc-top-app-bar__navigation-icon mdc-icon-button" aria-label="Open navigation menu">menu</button>
          <span class="mdc-top-app-bar__title">Page title</span>
        </section>
        <section class="mdc-top-app-bar__section mdc-top-app-bar__section--align-end" role="toolbar">
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Favorite">favorite</button>
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Search">search</button>
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Options">more_vert</button>
        </section>
      </div>
    </header>
    """
    tag = 'header'

    def __init__(self, *content, **context):
        action_items = context.pop('action_items', None)
        super().__init__(
            Div(
                # Navigation menu icon and page title.
                Component(
                    Button(
                        'menu',
                        aria_label="Open navigation menu",
                        cls='material-icons mdc-top-app-bar__navigation-icon mdc-icon-button',
                    ),
                    Span(
                        *content,
                        cls='mdc-top-app-bar__title',
                    ),
                    cls='mdc-top-app-bar__section mdc-top-app-bar__section--align-start',
                    tag='section',
                ),
                # Actions items and menu.
                MdcAppBarActions(action_items=action_items),
                cls='mdc-top-app-bar__row',
            ),
            cls='mdc-top-app-bar app-bar',
            id='app-bar',
            data_mdc_auto_init='MDCTopAppBar',
            **context,
        )


class MdcDrawer(Component):
    """
    <aside class="mdc-drawer mdc-drawer--dismissible">
      <div class="mdc-drawer__header">
        <h3 class="mdc-drawer__title">Drawer/h3>
        <h6 class="mdc-drawer__subtitle">Demonstration</h6>
      </div>
      <div class="mdc-drawer__content">
        <nav class="mdc-list">
          <a class="mdc-list-item mdc-list-item--activated" href="#" aria-current="page">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">inbox</i>
            <span class="mdc-list-item__text">Inbox</span>
          </a>
          <a class="mdc-list-item" href="#">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">send</i>
            <span class="mdc-list-item__text">Outgoing</span>
          </a>
          <hr class="mdc-list-divider">
          <h6 class="mdc-list-group__subheader">Labels</h6>
          <a class="mdc-list-item" href="#">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">drafts</i>
            <span class="mdc-list-item__text">Drafts</span>
          </a>
        </nav>
      </div>
    </aside>
    """
    tag = 'aside'

    def __init__(self, *content, **context):
        super().__init__(
            # Header.
            Div(
                H3(
                    *context['mdc-header-title'],
                    cls="mdc-drawer__title",
                ),
                H6(
                    *context['mdc-header-subtitle'],
                    cls="mdc-drawer__subtitle",
                ),
                cls="mdc-drawer__header",
            ) if 'mdc-header-title' in context else '',
            # List.
            Div(
                *content,
                cls='mdc-drawer__content',
            ),
            cls='mdc-drawer mdc-drawer--dismissible mdc-top-app-bar--fixed-adjust',
            data_mdc_auto_init='MDCDrawer',
            **context,
        )

class MdcAppContent(Div):
    """
    <div class="mdc-drawer-app-content mdc-top-app-bar--fixed-adjust">
      <main class="main-content" id="main-content">
        App Content
      </main>
    </div>
    """
    def __init__(self, *content, **context):
        super().__init__(
            Component(
                *content,
                tag='main',
                cls='main-content',
                id='main-content',
            ),
            cls='mdc-drawer-app-content mdc-top-app-bar--fixed-adjust',
        )


class MdcDrawerHeader(Div):
    """
    <div class="mdc-drawer__header">
      <h3 id="drawer-title" class="mdc-drawer__title">Title</h3>
      <h6 id="drawer-subtitle" class="mdc-drawer__subtitle">subtitle</h6>
    </div>
    """
    def __init__(self, *content, **context):
        title = context.pop('drawer_title', '')
        subtitle = context.pop('drawer_subtitle', '')
        super().__init__(
            *content,
            H3(
                title,
                cls='mdc-drawer__title',
                id='drawer-title',
            ),
            H6(
                subtitle,
                cls='mdc-drawer__subtitle',
                id='drawer-subtitle',
            ),
            cls='mdc-drawer__header',
        )


class MdcListDivider(Text):
    """
    <hr class="mdc-list-divider">
    <h6 class="mdc-list-group__subheader">Section</h6>
    """
    def __init__(self, *content, **context):
        super().__init__(
            Hr(cls="mdc-list-divider"),
            H6(
                *content,
                cls="mdc-list-group__subheader",
                **context,
            ),
        )


class MdcIcon(Component):
    tag = 'i'

    def __init__(self, *content, **context):
        super().__init__(
            *content,
            cls='material-icons mdc-list-item__graphic',
            aria_hidden='true',
            **context,
        )


class MdcList(Component):
    """ Allow tag to be overridden via context.

    """
    def __init__(self, *content, **context):
        context.setdefault('tag', 'ul')
        context.setdefault('cls', '')
        context['cls'] = (' '.join((context['cls'], 'mdc-list'))
                          if context['cls'] else 'mdc-list')
        super().__init__(*content, **context)


class MdcNavList(MdcList):
    def __init__(self, *content, **context):
        context['tag'] = 'nav'
        super().__init__(*content, **context)


class MdcListItem(Component):
    """ MdcListItem(Text("Home"), icon='home', href='#', active=True)
    <a class="mdc-list-item mdc-list-item--activated" href="#" aria-current="page">
      <span class="mdc-list-item__ripple"></span>
      <i class="material-icons mdc-list-item__graphic" aria-hidden="true">home</i>
      <span class="mdc-list-item__text">Home</span>
    </a>
    """
    def __init__(self, *content, icon=None, **context):
        context.setdefault('tag', 'a')
        context.setdefault('aria-current', 'page')
        context.setdefault('cls', '')
        context['cls'] = (' '.join((context['cls'], 'mdc-list-item'))
                          if context['cls'] else 'mdc-list-item')
        if context.pop('active', None):
            context['cls'] += ' mdc-list-item--activated'
            context['tabindex'] = 0
        super().__init__(
            Span(cls='mdc-list-item__ripple'),
            MdcIcon(icon) if icon else '',
            Span(*content, cls='mdc-list-item__text'),
            **context,
        )


class MdcAppBarActions(Component):
    def __init__(self, *content, **context):
        action_items = context.pop('action_items', None)
        if not action_items:
            return ''
        context['tag'] = 'section'
        super().__init__(
            *[Button(
                item[0],  # icon
                aria_label=item[1],  # label
                href=item[2],
                cls='material-icons mdc-top-app-bar__action-item mdc-icon-button',
            ) for item in action_items],
            cls='mdc-top-app-bar__section mdc-top-app-bar__section--align-end',
            role='toolbar',
            **context,
        )


class MdcButton(Div):
    def __init__(self, label=None, name=None, *content, **context):
        label = label or (name and name.capitalize())
        context.setdefault('tag', 'button')
        context.setdefault('cls', '')
        context['cls'] += 'mdc-button mdc-button--touch mdc-button--raised'
        super().__init__(
            # TODO: What tag here? Context?
            Component(
                Span(cls='mdc-button__ripple'),
                *content,
                Span(label, cls='mdc-button__label'),
                Span(cls='mdc-button__touch'),
                **context,
            ),
            cls='mdc-touch-target-wrapper',
            data_mdc_auto_init='MDCRipple',
        )


class MdcTextField(Label):
    """
    <label class="mdc-text-field mdc-text-field--filled mdc-text-field--label-floating">
      <span class="mdc-text-field__ripple"></span>
      <input class="mdc-text-field__input" type="text" aria-labelledby="my-label-id" value="Initial value">
      <span class="mdc-floating-label mdc-floating-label--float-above" id="my-label-id">
        Label in correct place
      </span>
      <span class="mdc-line-ripple"></span>
    </label>
    """
    def __init__(self, name=None, label=None, value=None, **context):
        label = label or (name and name.capitalize()) or ''
        super().__init__(
            Span(cls='mdc-text-field__ripple'),
            Input(
                name=name,
                value=value or '',
                type='text',
                aria_labelledby=f'id_{name}-label',
                cls='mdc-text-field__input',
                **context,
            ),
            Span(
                Text(label),
                id=f'{name}-label',
                cls='mdc-floating-label',
            ),
            Span(cls='mdc-line-ripple'),
            cls='mdc-text-field mdc-text-field--filled',
            data_mdc_auto_init='MDCTextField',
        )


class MDCLayoutGrid(Div):
    attrs = {'class': 'mdc-layout-grid'}


class MDCLayoutGridInner(Div):
    attrs = {'class': 'mdc-layout-grid__inner'}


class MDCLayoutGridCell(Div):
    attrs = {'class': 'mdc-layout-grid__cell'}



from ryzom.views import View

from .components.base import Document
from .components.home import Welcome
from .components.task import TaskList, TaskForm


class Layout(View):
    def oncreate(self, url):
        self.url = url
        self.content = Document(self)

    def onurl(self, url):
        self.url = url.strip('/')
        if self.url == 'todos':
            self.reactive('main-container', [TaskList(), TaskForm()])
        else:
            self.reactive('main-container', [Welcome()])
        return True

    def render(self, request):
        return self.content



import importlib
from django.conf import settings

from ryzom.request import Request


class RyzomMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        request.ryzom = None
        if getattr(settings, 'DDP_URLPATTERNS', None):
            urls = importlib.import_module(settings.DDP_URLPATTERNS).urlpatterns
            url = request.path_info.lstrip('/')
            for u in urls:
                if u.pattern.match(url):
                    request.ryzom_view = u.callback
                    break

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response



from ryzom_django.html import *
from ryzom_mdc import *
from ryzom_django.forms import field_kwargs, context_attrs, widget_attrs, widget_context


@template('django/forms/widgets/input.html')
@template('django/forms/widgets/date.html')
@template('django/forms/widgets/time.html')
@template('django/forms/widgets/text.html')
@template('django/forms/widgets/email.html')
@template('django/forms/widgets/password.html')
class MDCInputWidget(MDCTextFieldOutlined):
    @classmethod
    def from_boundfield(cls, bf):
        return cls(
            Input(**widget_attrs(bf)),
            label=bf.label,
            help_text=bf.help_text,
            errors=bf.errors,
        )


@template('django/forms/widgets/checkbox.html')
class MDCCheckboxWidget(MDCCheckboxField):
    @classmethod
    def from_boundfield(cls, bf):
        return cls(
            MDCCheckboxInput(**widget_attrs(bf)),
            **field_kwargs(bf),
        )


@template('django/forms/widgets/checkbox_select.html')
class MDCCheckboxSelectMultipleWidget(MDCCheckboxSelectField):
    @classmethod
    def from_boundfield(cls, bf):
        context = widget_context(bf)
        choices = []
        for group, options, index in context['optgroups']:
            if group is not None:
                continue
            choices += options

        return cls(
            Label(bf.label),
            *[
                Div(
                    MDCFormField(
                        MDCCheckboxInput(**context_attrs(choice)),
                        Label(choice['label'])
                    )
                )
                for choice in choices
            ],
            **field_kwargs(bf),
        )


@template('django/forms/widgets/multiwidget.html')
class MultiWidget(CList):
    def __init__(self, **context):
        super().__init__(*[
            templates[widget['template_name']](**widget)
            for widget in context['widget']['subwidgets']
        ])

    @classmethod
    def from_boundfield(cls, bf):
        return Div(
            Label(bf.label),
            cls(**widget_context((bf))),
            cls='mdc-form-field'
        )


@template('django/forms/widgets/splitdatetime.html')
class SplitDateTimeWidget(MDCField):
    @classmethod
    def from_boundfield(cls, bf):
        context = widget_context(bf)
        style = 'margin-bottom: 0; margin-top: 12px'
        return cls(
            Label(bf.label),
            MDCFormField(
                MDCTextFieldOutlined(
                    Input(**context_attrs(context['subwidgets'][0])),
                    label='Date',
                    style='margin-right: 12px; ' + style,
                ),
                MDCTextFieldOutlined(
                    Input(**context_attrs(context['subwidgets'][1])),
                    label='Time',
                    style=style,
                ),
            ),
            **field_kwargs(bf),
        )


@template('django/forms/widgets/textarea.html')
class TextareaWidget(MDCTextareaFieldOutlined):
    @classmethod
    def from_boundfield(cls, bf):
        attrs = widget_attrs(bf)
        return cls(
            Textarea(
                attrs.pop('value', '') or '',
                aria_labelledby=f'id_{bf.html_name}_label',
                **attrs,
            ),
            label=bf.label,
            help_text=bf.help_text,
            errors=bf.errors,
        )



from django import forms
from django.urls import path
from django.views import generic

from ryzom.js.renderer import JS
import ryzom_mdc as html


class ExampleDocument(html.Html):
    stylesheets = [
        'https://fonts.googleapis.com/icon?family=Material+Icons',
        'https://fonts.googleapis.com/css2?family=Nanum+Pen+Script&display=swap',
        'https://unpkg.com/material-components-web@latest/dist/material-components-web.min.css',
    ]
    scripts = [
        'https://unpkg.com/material-components-web@latest/dist/material-components-web.min.js',
        'static/ryzom/js/py-builtins.js',
    ]

    def __init__(self, *content, **context):
        links = [
            html.Link(href=src, rel='stylesheet')
            for src in self.stylesheets
        ]
        scripts = [
            html.Script(src=src, type='text/javascript')
            for src in self.scripts
        ]
        scripts.append(html.Script('mdc.autoInit()', type='text/javascript'))

        body = html.Body(
            *content,
            cls='mdc-typography',
        )

        scripts.append(html.Script(body.render_js_tree(), type='text/javascript'))

        body.addchildren(scripts)

        super().__init__(
            html.Head(
                html.Meta(charset='utf-8'),
                html.Meta(
                    name='viewport',
                    content='width=device-width, initial-scale=1.0',
                ),
                html.Title('Secure elections with homomorphic encryption'),
                *links,
            ),
            body
        )


# Serves to demonstrate template composition based on multi level nesting
class ExampleCard(html.Div):
    def __init__(self, *content, **context):
        super().__init__(*content, style='max-width: 20em; margin: auto')


# in general, you wouldn't go through a template_name, but since this is useful
# to design views that you import from external packages, we have this example
# here, it also shows how you can compose by nesting different layout objects
@html.template('form.html', ExampleDocument, ExampleCard)
class ExampleFormViewComponent(html.Html):
    def __init__(self, *content, view, form, **context):
        # view and form come from the default context, we're spreading them as
        # nice, required variables for this template.

        # As you can imagine, having templates in Python not only gives you all
        # the programing power, but it also allows you to use a debugger
        # breakpoint() which was not possible with traditionnal templates.

        content = []

        if view.request.method == 'POST' and form.is_valid():
            content += [
                html.Div(
                    html.H3('Form post success!'),
                    html.MDCList(*[
                        html.MDCListItem(f'{key}: {value}')
                        for key, value in form.cleaned_data.items()
                    ])
                )
            ]

        content.append(
            html.Form(
                html.CSRFInput(view.request),
                form,
                method="post",
            ),
        )
        super().__init__(*content)


class ExampleForm(forms.Form):
    char = forms.CharField(
        help_text='Non required example text input',
    )
    email = forms.EmailField(
        help_text='Valid email required',
    )
    boolean = forms.BooleanField(
        help_text='Required boolean!',
    )
    checkboxes = forms.MultipleChoiceField(
        choices=(('a', 'a'), ('b', 'b')),
        widget=forms.CheckboxSelectMultiple,
        help_text='Required checkbox multiple',
    )
    datetime = forms.SplitDateTimeField(
        widget=forms.SplitDateTimeWidget(
            date_attrs=dict(type='date'),
            time_attrs=dict(type='time'),
        ),
        help_text='Required help text',
    )
    textarea = forms.CharField(
        widget=forms.Textarea,
        help_text='This is the help text'
    )

    # Let's override the default rendering to add a submit button
    def to_component(self):
        class FormDiv(html.Div):
            def render_js(self):
                def setvar():
                    setattr(window, 'toto', 'toto')

                return JS(setvar)

        return FormDiv(
            html.H3('Example form!'),
            super().to_component(),
            html.Div(
                html.MDCButtonOutlined('Submit'),
            )
        )


# Finally, a Django FormView, there's nothing to see here because of how well
# Ryzom integrates with Django. Of course you're free to make views that do
# some crazy Ryzom rendering, this merely shows how you would hook in the
# default Django rendering on external views that you include and do not want
# to fork: you can just override the default template with @html.template
# instead of by creating html templates.
class ExampleFormView(generic.FormView):
    template_name = 'form.html'
    form_class = ExampleForm

    def form_valid(self, form):
        # we don't have a success url, render again on form_valid
        return super().get(self.request)


urlpatterns = [
    path('', ExampleFormView.as_view())
]



'''
Project URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

""" original `project/urls.py`
from django.conf.urls import include, url

urlpatterns = [
    url('^', include('ryzom.urls')),
]
'''
from crudlfap import shortcuts as crudlfap

from django.conf import settings
from django.conf.urls import include, url, re_path
from django.utils.translation import ugettext_lazy as _
from django.views.generic import RedirectView
from django.views.static import serve

from .views import Home


crudlfap.site.title = _('Ryzom - Demo')  # used by base.html
crudlfap.site.urlpath = 'admin'  # example url prefix
crudlfap.site.views['home'] = Home

urlpatterns = [
    crudlfap.site.urlpattern,
    url(r'^favicon\.ico$', RedirectView.as_view(url='/static/images/favicon.ico')),
    re_path(
        r'^%s(?P<path>.*)$' % settings.STATIC_URL.lstrip('/'),
        serve,
        kwargs=dict(document_root=settings.STATIC_ROOT)
    ),
    url('', include('ryzom.urls')),
]
if 'debug_toolbar' in settings.INSTALLED_APPS and settings.DEBUG:
    import debug_toolbar
    urlpatterns += [
        url(r'^__debug__/', include(debug_toolbar.urls)),
    ]



import os
from pathlib import Path
BASE_DIR = Path(__file__).resolve().parent.parent
SECRET_KEY = '4am4pn_87&v0qaq%_-2me06et#@prq(yp6npk8g495!@7s1hoi'
DEBUG = True
ALLOWED_HOSTS = []
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    # Enable components templates auto discover
    'ryzom_django',
    'ryzom',
    # Enable Reactive components models
    # 'ryzom_django.apps.ReactiveConfig',

    # Enable form rendering with MDC components
    'ryzom_django_mdc',

    # Enable default templates too, until we implement them all
    'django.forms',
]

MIDDLEWARE = [
    # Enable Reactive middleware
    'ryzom_django.middleware.RyzomMiddleware',

    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

# Enable Ryzom form rendering
FORM_RENDERER = 'django.forms.renderers.TemplatesSetting'

# Enable Ryzom template backend
TEMPLATES = [
    {
        "BACKEND": "ryzom_django.template_backend.Ryzom",
        "OPTIONS": {
            "app_dirname": "components",
            # "components_module": "ryzom.components.muicss",
            # "components_prefix": "Mui",
            # "components_module": "ryzom.components.django",
            # "components_prefix": "Django",
            # "context_processors": [
            #     "django.template.context_processors.debug",
            #     "django.template.context_processors.request",
            #     "django.contrib.auth.context_processors.auth",
            #     "django.contrib.messages.context_processors.messages",
            #     "django.template.context_processors.i18n",
            # ],
            # "autoescape": True,
            # "auto_reload": DEBUG,
            # "translation_engine": "django.utils.translation",
            # "debug": False,
        }
    },
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

ROOT_URLCONF = 'ryzom_django_example.urls'

WSGI_APPLICATION = 'ryzom_django_example.wsgi.application'

DATABASES = {
    'default': {
        'ENGINE': os.getenv('DB_ENGINE', 'django.db.backends.postgresql'),
        'NAME': os.getenv('DB_NAME', 'ryzom_django_example'),
        'HOST': os.getenv('DB_HOST'),
        'USER': os.getenv('DB_USER'),
        'PASSWORD': os.getenv('DB_PASSWORD'),
    }
}


# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = BASE_DIR / 'static'
